/***************************************************************************
begin                : Thu Apr 24 15:54:58 CEST 2003
copyright            : (C) 2003 by Giuseppe Lipari
email                : lipari@sssup.it
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cpu.hpp>
#include <assert.h>
#include <energyMRTKernel.hpp>

namespace RTSim
{

    unsigned int CPU_BL::referenceFrequency;

    string Island_BL::IslandName[NUM_ISLANDS] = {"BIG","LITTLE"};

    CPU::CPU(const string &name, const vector<volt_type> &V,
             const vector<freq_type> &F, CPUModel *pm)
        : Entity(name), frequencySwitching(0), index(0) {
        auto num_OPPs = V.size();

        cpuName = name;

        if (num_OPPs == 0) {
            PowerSaving = false;
            return;
        }

        PowerSaving = true;

        // Setting voltages and frequencies
        for (int i = 0; i < num_OPPs; i ++)
        {
            OPP opp;
            opp.voltage = V[i];
            opp.frequency = F[i];
            OPPs.push_back(opp);
        }

        // Creating the Energy Model class
        // and initialize it with the max values
        if (!pm) {
            powmod = new CPUModelMinimal(OPPs[currentOPP].voltage, OPPs[currentOPP].frequency);
            powmod->setFrequencyMax(OPPs[OPPs.size()-1].frequency);
        } else {
            powmod = pm;
            pm->setCPU(this);
        }

        /* Use the maximum OPP by default */
        currentOPP = num_OPPs - 1;

        _workload = "idle";

        // Setting speeds (basing upon frequencies)
        for (unsigned int opp = 0; opp < OPPs.size(); ++opp)
            speeds[opp] = getSpeedByOPP(opp);
    }

    CPU::~CPU()
    {
        OPPs.clear();
    }

    int CPU::getOPP()
    {
        if (PowerSaving)
            return currentOPP;
        return 0;
    }

    void CPU::setOPP(unsigned int newOPP)
    {
        //std::cout << __func__ << " setting currentOPP from " << currentOPP << " to " << newOPP << ", OPPs.size()=" << OPPs.size() << std::endl;
        assert(newOPP < OPPs.size() && newOPP >= 0);
        currentOPP = newOPP;
        updateCPUModel();
    }

    freq_type CPU::getFrequency() const {
        return OPPs[currentOPP].frequency;
    }

    volt_type CPU::getVoltage() const {
        return OPPs[currentOPP].voltage;
    }

    watt_type CPU::getMaxPowerConsumption() {
        return _max_power_consumption;
    }

    void CPU::setMaxPowerConsumption(watt_type max_p) {
        _max_power_consumption = max_p;
    }

    watt_type CPU::getCurrentPowerConsumption() {
        if (PowerSaving) {
            updateCPUModel();
            return (powmod->getPower());
        }
        return 0.0;
    }

    watt_type CPU::getCurrentPowerSaving() {
        if (PowerSaving)
        {
            long double maxPowerConsumption = getMaxPowerConsumption();
            long double saved = maxPowerConsumption - getCurrentPowerConsumption();
            return static_cast<watt_type>(saved / maxPowerConsumption);
        }
        return 0;
    }

    speed_type CPU::setSpeed(double newLoad) {
        DBGENTER(_KERNEL_DBG_LEV);
        DBGPRINT("pwr: setting speed in CPU::setSpeed()");
        DBGPRINT("pwr: New load is " << newLoad);
        if (PowerSaving)
        {
            DBGPRINT("pwr: PowerSaving=on");
            DBGPRINT("pwr: currentOPP=" << currentOPP);
            for (int i=0; i < (int) OPPs.size(); i++)
                if (speeds[i] >= newLoad) {
                    if (i != currentOPP)
                        frequencySwitching++;
                    currentOPP = i;
                    DBGPRINT("pwr: New OPP=" << currentOPP << " New Speed="
                                             << speeds[currentOPP]);

                    return speeds[i]; // It returns the new speed
                }
        }
        else
            DBGPRINT("pwr: PowerSaving=off => Can't set a new speed!");

        return 1; // An error occurred or PowerSaving is not enabled
    }

    void CPU::setWorkload(const string &workload)
    {
	if (workload == "idle")
		DBGPRINT("\t" << getName() << " gets wl idle at t=" << SIMUL.getTime());
        _workload = workload;
    }

    string CPU::getWorkload() const
    {
        return _workload;
    }

    speed_type CPU::getSpeed() {
        if (PowerSaving) {
            updateCPUModel();
            return powmod->getSpeed();
        }
        return 1.0;
    }

    speed_type CPU::getSpeedByOPP(unsigned int opp) {
        if (!PowerSaving)
            return 1;
        assert(opp < OPPs.size() && opp >= 0);
        int old_curr_opp = currentOPP;
        setOPP(opp);
        double s = getSpeed();
        setOPP(old_curr_opp);
        return s;
    }

    unsigned long int CPU::getFrequencySwitching()
    {
        DBGENTER(_KERNEL_DBG_LEV);
        DBGPRINT("frequencySwitching=" << frequencySwitching);

        return frequencySwitching;
    }

    void CPU::check()
    {

#define __get_speed(iter) (speeds.begin() + (iter - OPPs.begin()))
        cout << "Checking CPU:" << cpuName << endl;;
        cout << "Max Power Consumption is :" << getMaxPowerConsumption() << endl;
        for (vector<OPP>::iterator iter = OPPs.begin(); iter != OPPs.end(); iter++)
        {
            cout << "-OPP-" << endl;
            cout << "\tFrequency:" << (*iter).frequency << endl;
            cout << "\tVoltage:" << (*iter).voltage << endl;
            cout << "\tSpeed:" << *__get_speed(iter) << endl;
        }
        for (unsigned int i = 0; i < OPPs.size(); i++)
            cout << "Speed level" << getSpeedByOPP(i) << endl;
        for (vector<OPP>::iterator iter = OPPs.begin(); iter != OPPs.end(); iter++)
        {
            cout << "Setting speed to " << *__get_speed(iter) << endl;
            setSpeed(*__get_speed(iter));
            cout << "New speed is  " << getSpeed() << endl;
            cout << "Current OPP is  " << getOPP() << endl;
            cout << "Current Power Consumption is  " << getCurrentPowerConsumption() << endl;
            cout << "Current Power Saving is  " << getCurrentPowerSaving() << endl;
        }
    }

    uniformCPUFactory::uniformCPUFactory()
    {
        _curr=0;
        _n=0;
        index = 0;
    }

    uniformCPUFactory::uniformCPUFactory(char* names[], int n)
    {
        _n=n;
        _names = new char*[n];
        for (int i=0; i<n; i++)
        {
            _names[i]=names[i];
        }
        _curr=0;
        index = 0;
    }

    CPU *uniformCPUFactory::createCPU(const string &name,
                                      const vector<volt_type> &V,
                                      const vector<freq_type> &F,
                                      CPUModel *pm) {
        CPU *c;

        if (_curr==_n)
            c = new CPU(name, V, F, pm);
        else
            c = new CPU(_names[_curr++], V, F, pm);

        c->setIndex(index++);
        return c;
    }

    /// to string operator
    ostream& operator<<(ostream &strm, CPU &a) {
      return strm << a.toString();
    }

    // useless, because entity are copied and thus they change name by implementation of Entity...
    bool operator==(const CPU& lhs, const CPU& rhs) {
      return lhs.getName() == rhs.getName();
    }

    void CPU::updateCPUModel() {
        powmod->setVoltage(getVoltage());
        powmod->setFrequency(getFrequency());
    }

    // ------------------------------------------------------------- big little
    freq_type CPU_BL::getFrequency() const {
        return _island->getFrequency();
    }

    volt_type CPU_BL::getVoltage() const {
        return _island->getVoltage();
    }

    unsigned int CPU_BL::getOPP() const {
        return _island->getOPP();
    }

    void CPU_BL::setOPP(unsigned int opp) {
        _island->setOPP(opp);
    }

    vector<struct OPP> CPU_BL::getHigherOPPs() {
        return _island->getHigherOPPs();
    }

    void CPU_BL::setBusy(bool busy) {
        _isBusy = busy;
    }

    bool CPU_BL::isIslandBusy() {
        return _island->isBusy();
    }

    IslandType CPU_BL::getIslandType() {
        return _island->getIslandType();
    }

    freq_type CPU_BL::getFrequency(unsigned int opp) const {
        return _island->getFrequency(opp);
    }

    watt_type CPU_BL::getPowerConsumption(freq_type frequency) {
        // Find what OPP corresponds to provided frequency
        unsigned int old_opp = getOPP();
        unsigned int opp = _island->getOPPByFrequency(frequency);
        setOPP(opp);
        double pow = getCurrentPowerConsumption();
        setOPP(old_opp);
        return pow;
    }

    speed_type CPU_BL::getSpeed(freq_type freq) {
        unsigned int opp = _island->getOPPByFrequency(freq);
        return getSpeedByOPP(opp);
    }

    speed_type CPU_BL::getSpeedByOPP(unsigned int opp) {
        int old_curr_opp = getOPP();
        setOPP(opp);
        speed_type s = getSpeed();
        setOPP(old_curr_opp);
        return s;
    }

    speed_type CPU_BL::getSpeed() {
        assert(getOPP() < _island->getOPPsize() && getOPP() >= 0);
        updateCPUModel();
        return _pm->getSpeed();
    }

    void CPU_BL::updateCPUModel() {
        _pm->setVoltage(getVoltage());
        _pm->setFrequency(getFrequency());
    }


    void Island_BL::setOPP(unsigned int opp) {
        assert(opp >= 0 && opp < _opps.size());
        _currentOPP = opp;
        for (CPU_BL* c : getProcessors())
            c->updateCPUModel();
        _kernel->onOppChanged(_currentOPP, this);
    }

}
